
/* NetPercept.C
 * SBC++ (soccer client for Robocup2000)
 * Eslam Nazemi<nazemi@cc.sbu.ac.ir>
 * Amin Abbaspour<a-abbaspour@ce.sbu.ac.ir>
 * Mahmood Rahmani<m-rahmani@ce.sbu.ac.ir>
 * Bahman Radjabalipoor<b-radjabalipoor@ce.sbu.ac.ir>
 * Computer Department
 * Shahid Beheshti University
*/

#include "NetFuncs.h"
#include "NetHeader.h"
#include "client.h"
#include "pass.h"

#define MAX_IMAGINE_TIME	30	
#define MIN_CONGESTION		0.001
#define MAX_CONGESTION		0.1

/*-----------------------------------------------------------------------------
Function: float TNetCommand::GetLinearFuzzyValue(float value, float left, float right, int up)
 Date    : 13.03.98
-----------------------------------------------------------------------------*/
float GetLinearFuzzyValue(const float& value, const float& left, 
					const float& right, const int& up)
{
    if (up)
    	if (value < left)  return 0.0;
    	else if (value > right) return 1.0;
	    else return (value - left) / (right - left);
    else
	if (value < left) return 1.0;
	else if (value > right)	return 0.0;
	    else return (right - value) / (right - left);
}


/*-----------------------------------------------------------------------------
Function: Perception functions
Date    : 14.01.98
-------------------------------------------------------------------------------
Parameter: do not get parameters
Returnval: activation (truth) of this perception
Description: this are the functions that calculate the truth of perceptions
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
just nothing.
-----------------------------------------------------------------------------*/
float GetNothing()						/* OK */
{
    static float val = 0.1;

    Mem->LogAction3(200,"Precp Nothing  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
returns true, if the ball is nearer to the other goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
float GetBallAhead()						/* OK */
{
    static float val = 0.0;

    int x1 = (int) Mem->MyX();
    int x2 = (int) Mem->BallX();

    if ((x2 < 47) && (x2 > x1) && Mem->BallPositionValid())
	val = GetLinearFuzzyValue(x2 - x1, 7.0, 20.0);
    else
	val = 0.0;
	
    Mem->LogAction3(200,"Precp Ball-Ahead  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
returns true, if the ball is nearer to the own goal (only x coordinate)
than this player is
-----------------------------------------------------------------------------*/
float GetBallBehind()						/* OK */
{
    static float val = 0.0;

    int x1 = (int) Mem->MyX();
    int x2 = (int) Mem->BallX();

    if (x2 > -47 && Mem->BallPositionValid())  
	val = GetLinearFuzzyValue(x1 - x2, -4.0, 0.0);
    else
	val = 0.0;
	
    Mem->LogAction3(200,"Precp Ball-Behind  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
calculates if the ball is or will be within reach to catch (goalie)
-----------------------------------------------------------------------------*/
float GetBallIsCatchable()					/* OK */
{
    static float val = 0.0;

    val = GetJustCaughtBall() ? 0 : Mem->BallCatchable();

    Mem->LogAction3(200,"Precp Ball-Is-Catchable  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
false up to 15 m of ball distance to agent, true at 25 or greater distance
-----------------------------------------------------------------------------*/
float GetBallIsFar()						/* OK */
{
    static float val = 0.0;

    if (Mem->BallPositionValid())
	val = GetLinearFuzzyValue(Mem->BallDistance(), 15.0, 25.0);
    else
	val = 0.0;
	
    Mem->LogAction3(200,"Precp Ball-Is-Far  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
perception proposition for the distance of the ball to the other goal
-----------------------------------------------------------------------------*/
float GetBallIsNearOtherGoal()					/* OK */
{
    static float val = 0.0;

    if(Mem->BallPositionValid())
	val = GetLinearFuzzyValue(
		Mem->BallAbsolutePosition().dist(Mem->MarkerPosition( Mem->RM_Their_Goal )), 
                   15.0, 50.0, FALSE);
    else
	val = 0.0;				     

    Mem->LogAction3(200,"Precp Ball-Is-Near-Other-Goal  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
perception proposition for the distance of the ball to the own goal
-----------------------------------------------------------------------------*/
float GetBallIsNearOwnGoal()					/* OK */
{
    static float val = 0.0;

    if(Mem->BallPositionValid())
	val = GetLinearFuzzyValue(
		Mem->BallAbsolutePosition().dist(Mem->MarkerPosition( Mem->RM_My_Goal )),
			             15.0, 50.0, FALSE);
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Ball-Is-Near-Own-Goal  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
perception proposition for the (future) distance of the ball to the own goal
-----------------------------------------------------------------------------*/
float GetBallIsVeryNearOwnGoal()				/* OK */
{
    static float val = 0.0;

    Vector OwnGoal = Mem->MarkerPosition( Mem->RM_My_Goal );
    if(Mem->BallPositionValid())
	val = GetLinearFuzzyValue(
		    OwnGoal.dist(Mem->BallPredictedPosition(10)),
15.0, 25.0, FALSE)
			+ (Mem->BallInOwnPenaltyArea() ? 0.3 : 0);
    else
	val = 0.0;
	
    val =  Min(1.0, val );

    Mem->LogAction3(200,"Precp Ball-Is-Very-Near-Own-Goal  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the ball is within the kickable_area of this agent
-----------------------------------------------------------------------------*/
float GetBallKickable()						/* OK */
{
    static float val = 0.0;

    if(Mem->BallKickable()) 
	val = 1.0;
    else {
	Vector intersection = Mem->BallAbsolutePosition();
	
	if(Mem->BallWillBeKickable(2)) 
	    intersection = Mem->BallPredictedPosition(2);
	else if(Mem->BallWillBeKickable(1)) 
	    intersection = Mem->BallPredictedPosition(1);

	val = GetLinearFuzzyValue(Mem->DistanceTo(intersection), 
	    	    		    Mem->SP_kickable_area, 1.4, FALSE) * GetHaveSeenBall();
    }
    
    Mem->LogAction3(200,"Precp Ball-Kickable  \t\t%1.2f", val);
    return val;
}


/*-----------------------------------------------------------------------------
true at kick_in or corner_kick
-----------------------------------------------------------------------------*/
float GetBallOutside()						/* OK */
{
    static float val = 0.0;

    if ( Mem->PlayMode == PM_My_Kick_In || 
Mem->PlayMode == PM_My_Corner_Kick ) 
    	val = 1.0;
    else 
    	val = 0.0;

    Mem->LogAction3(200,"Precp Ball-OutSide  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, immediately after a successfull catch
-----------------------------------------------------------------------------*/
float GetCatchSuccess()		
{
    static float val = 0.0;

    if (Mem->PlayMode == PM_My_Goalie_Free_Kick  && 
	    ( Mem->CurrentTime.t - Mem->LastCatchTime.t ) < 2 )
	val = 1.0;
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Catch-Success  \t\t%1.2f", val);
    return val;
}
/*-----------------------------------------------------------------------------
true, if the free teammate is in offside
-----------------------------------------------------------------------------*/
float GetFreeTeammateIsOffside()
{
    static float val = 0.0;

    //Unum pTeammate = CalculateFreestTeammate();
    
    Unum pTeammate = Mem->FurthestForwardTeammate();
    	
    if ( pTeammate != Unum_Unknown )
 
	val = Mem->TeammateInOffsidePosition(pTeammate);
    else
	val = 1.0;

    Mem->LogAction3(200,"Precp Teammate-Is-Offside \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, if this agent is nearest to ball and near enough (but maybe farer than ballKickable)
-----------------------------------------------------------------------------*/
float GetHaveBall()			
{
    static float val = 0.0;

    float ballDist = Mem->BallDistance();
    Unum  player = Mem->PlayerWithBall(); 

    if(player != Mem->MyNumber) 
	val = 0.0;
    else
	val = GetLinearFuzzyValue(ballDist, Mem->SP_kickable_area, 4.0, FALSE) * GetHaveSeenBall();

    Mem->LogAction3(200,"Precp Have-Ball  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
is opponet marked
-----------------------------------------------------------------------------*/
float GetHaveMarked()		
{
    static float val = 0.0;

    Mem->LogAction3(200,"Precp Have-Marked  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, if this agent has enough stamina to dash 100% to the ball
-----------------------------------------------------------------------------*/
float GetHaveEnoughStamina()			
{
    static float val = 0.0;

    float ballDist = Mem->BallPositionValid() ? Mem->BallDistance() : 50;
    float stamina  = Mem->MyStamina();

    val = GetLinearFuzzyValue(stamina - ballDist * 85, 0.0, 
				Mem->SP_effort_dec_thr * Mem->SP_stamina_max);

    Mem->LogAction3(200,"Precp Have-Enough-Stamina  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, if the ball has been seen within the last 3 seconds
-----------------------------------------------------------------------------*/
float GetHaveSeenBall()					
{
    static float val = 0.0;
    int deltaTime = Mem->CurrentTime.t - Mem->GetBall()->GetSeenTime().t;

    if(!Mem->BallPositionValid())
	val = 0.0;
    else		
	val = GetLinearFuzzyValue( deltaTime, 0.0, 
MAX_IMAGINE_TIME, FALSE);	
    
    Mem->LogAction3(200,"Precp Have-Seen-Ball  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
the same as GetLowStamina just another name.
-----------------------------------------------------------------------------*/
float GetLoseStamina()						
{
    static float val = 0.0;

    val =  1.0 - GetLowStamina();

    Mem->LogAction3(200,"Precp Lose-Stamina  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the player has got enough power to dash 
-----------------------------------------------------------------------------*/
float GetHighStamina()				
{
    static float val = 0.0;

    val = GetLinearFuzzyValue(Mem->MyStamina(), 
1500.0,
 Mem->SP_stamina_max);

    Mem->LogAction3(200,"Precp High-Stamina  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, if the agent knows he is in offside
-----------------------------------------------------------------------------*/
float GetInOffside()		
{
    static float val = 0.0;

    val = Mem->InOffsidePosition();

    Mem->LogAction3(200,"Precp In-Offside  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the agent is in the opponents half
-----------------------------------------------------------------------------*/
float GetInOtherHalf()					
{
    static float val = 0.0;

    char Location = Mem->MyX() > 0 ? 'r' : 'l';
    if ( Location == Mem->MySide )
	val = 0.0;
    else
	val = 1.0;

    Mem->LogAction3(200,"Precp In-Other-Half  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, if the goalie caught the ball within the last 5 cycles. A goalie can only
catch the ball once each 5 cycles
-----------------------------------------------------------------------------*/
float GetJustCaughtBall()			
{
    static float val = 0.0;

    if( (Mem->CurrentTime.t - Mem->LastCatchTime.t) < 5 ) 
	val = 1.0;
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Just-Caught-Ball  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
if i kown which opponent to mark
-----------------------------------------------------------------------------*/
float GetKnowMarkedOpponent()				/* not used */
{
    static float val = 0.0;

    if(Mem->MarkedOpponent > 0 && Mem->MarkedOpponent <= Mem->SP_team_size)
	val = 1.0;
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Know-Markded-Opponent  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the player knows he is not in offside
-----------------------------------------------------------------------------*/
float GetKnowNotInOffside()				/* not used */
{
    static float val = 0.0;
    Unum opp = Mem->FurthestBackOpponent();

    if(opp == Unum_Unknown || (Mem->MyX() > Mem->OpponentX(opp) && Mem->BallPossessor() != Mem->MyNumber))
	val = 1.0;
    else
	val = 0.0;
	
    Mem->LogAction3(200,"Precp Know-Not-In-Offside  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the player has got low power to dash 
-----------------------------------------------------------------------------*/
float GetLowStamina()			
{
    static float val = 0.0;
    float iStart = Mem->SP_effort_dec_thr * Mem->SP_stamina_max;
    int	  iTime = Mem->LastSenseTime.t;

    if ((iTime > 5500) && (iTime < 6000))
 iTime -= 2000;
    if ((iTime > 2500) && (iTime < 2000))
 iStart = 500;

    val = GetLinearFuzzyValue(Mem->MyStamina(), iStart, iStart + 200, FALSE);

    Mem->LogAction3(200,"Precp Low-Stamina  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the player is on turn to say something
-----------------------------------------------------------------------------*/
float GetMaySaySomething()					
{
    static float val = 0.0;

    int	time = Mem->LastSenseTime.t;
    if ((!Mem->ClockStopped) && !(time % 2) 
		&& ( ( (time * 2) % 11) + 1 == Mem->MyNumber))
        val = 1.0;
    else
	val = 0.0;
    
    Mem->LogAction3(200,"Precp May-Say-Something  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if this player is close to the ball
-----------------------------------------------------------------------------*/
float GetNearBall()				
{
    static float val = 0.0;

    val = GetLinearFuzzyValue(Mem->BallDistance(), 5.0, 25.0, FALSE)
 * GetHaveSeenBall();

    Mem->LogAction3(200,"Precp Near-Ball  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if this player is close to the goal
-----------------------------------------------------------------------------*/
float GetNearGoal()					
{
    static float val = 0.0;

    val = GetLinearFuzzyValue(Mem->DistanceTo(Mem->MarkerPosition( Mem->RM_Their_Goal )), 10.0, 25.0, FALSE);

    Mem->LogAction3(200,"Precp Near-Goal  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if this player is close to an opponent
-----------------------------------------------------------------------------*/
float GetNearOpponent()				
{
    static float val = 0.0;

    Unum opponent = Mem->ClosestOpponent();
    if (opponent == Unum_Unknown) val = 0.0;
    else	
        val = GetLinearFuzzyValue(Mem->OpponentDistance(opponent)
, 1.0, 6.0, FALSE);

    Mem->LogAction3(200,"Precp Near-Opponent  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if this player is close to the own goal
-----------------------------------------------------------------------------*/
float GetNearOwnGoal()			
{
    static float val = 0.0;

    val = GetLinearFuzzyValue(Mem->DistanceTo(Mem->MarkerPosition( Mem->RM_My_Goal )), 10.0, 30.0, FALSE);

    Mem->LogAction3(200,"Precp Near-Own-Goal  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
only false in the case that the opponent scored a goal
-----------------------------------------------------------------------------*/
float GetNoOtherGoal()					
{
    static float val = 0.0;

    if ( Mem->PlayMode == PM_My_Free_Kick )
 val = 1.0;
    else
 val = 0.0;

    Mem->LogAction3(200,"Precp No-Other-Goal  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
just for goal offsideTrap 
-----------------------------------------------------------------------------*/
float GetOffsideTrap()	
{
    static float val = 0.0;

    Mem->LogAction3(200,"Precp Offside-Trap  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if opponent with ball is free
-----------------------------------------------------------------------------*/
float GetOpponentIsFree()
{
    static float val = 0.0;

    Unum opponent = Mem->BallPossessor();
    if(opponent > 0 ) 	val = 0.0;
    else
	if(Mem->OpponentDistanceTo(opponent, Mem->MarkerPosition(Mem->RM_My_Goal)) > 25.0)
	    val =  GetLinearFuzzyValue(Mem->OpponentCongestionRelToTeammate(opponent),
					0.0,0.8,FALSE);
	else
	    val = 0.0;
	    
    val = Max(GetLinearFuzzyValue(Mem->OpponentCongestionRelToTeammate(opponent), 0.0, 0.8, FALSE), 0.0);

    Mem->LogAction3(200,"Precp Opponent-Is-Free  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if an opponent is between this agent and the opponents goal
-----------------------------------------------------------------------------*/

float GetOpponentInFront()		
{
    static float val = 0.0;

    AngleDeg Ang	= 45;
    AngleDeg ofAng	= Mem->AngleToFromBody( Mem->GetMyRunWithBallTarget() );
    float Dist		= 5;
    float congestion    = 0;
    
float dist2;
	
    for (int i = 1; i <= Mem->SP_team_size; i++)
 {
	if ( Mem->OpponentPositionValid(i) &&
 !Mem->IsOpponentBehind(i, Mem->MyPos()) &&
		Mem->OpponentDistance(i) < Dist &&
		fabs(GetNormalizeAngleDeg(Mem->OpponentAngleFromBody(i) - ofAng)) < Ang)
 {
		dist2 = Mem->OpponentDistance2To(i, Mem->MyPos()) + 0.1;
		congestion += 1/dist2;
	}
	if (congestion > MAX_CONGESTION)
		break;
    }

    val = GetLinearFuzzyValue(congestion, MIN_CONGESTION, MAX_CONGESTION);

    Mem->LogAction3(200,"Precp Opponent-InFront  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if playmode is play_on
-----------------------------------------------------------------------------*/
float GetPlayModeOn()				
{
    static float val = 0.0;

    val = 0.0;
    if(Mem->PlayMode == PM_My_Goal_Kick && Mem->CP_goalie) val = 1.0;
    else 
    if (Mem->PlayMode == PM_Play_On || Mem->PlayMode == PM_My_Kick_Off ||
    	Mem->PlayMode == PM_My_Offside_Kick || Mem->PlayMode == PM_My_Kick_In ||
	Mem->PlayMode == PM_My_Corner_Kick ) val = 1.0;

    Mem->LogAction3(200,"Precp Play-Mode-On  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if a goal has been scored this cycle
-----------------------------------------------------------------------------*/
float GetSoccerGoal()						
{
    static float val = 0.0;

    if ( Mem->PlayMode == PM_Their_Free_Kick )
	val = 1.0;
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Soccer-Goal  \t\t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true, if this agent is unmarked i.e. the ball can be passed to him
-----------------------------------------------------------------------------*/
float GetStandFree()				
{
    static float val = 0.0;

    Unum opponent = Mem->ClosestOpponent();
    if (opponent == Unum_Unknown)
    	val = 1.0;
    else	
	val = GetLinearFuzzyValue(Mem->OpponentDistance(opponent)
, 3.0, 8.0);

    Mem->LogAction3(200,"Precp Stand-Free  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if a teammate is behind me in x direction (used for offside trap)
-----------------------------------------------------------------------------*/
float GetTeammateBehind()			
{
    static float val = 0.0;
    static float  fTeammateX;
    
    Unum pPlayer = Mem->FurthestBackTeammate();
    if (pPlayer != Unum_Unknown)
 {
	fTeammateX =  Mem->MyX() - Mem->TeammateX(pPlayer);
	val = GetLinearFuzzyValue(fTeammateX, 4.0, 9.0);
    }
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Teammate-Behind  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if the teammate is the closest player to the ball and is close enough
-----------------------------------------------------------------------------*/
float GetTeammateHasBall()				
{
    static float val = 0.0;

    Unum pPlayer = Mem->BallPossessor();
    if ((pPlayer > 0) && (pPlayer != Mem->MyNumber))
	val = GetLinearFuzzyValue(Mem->TeammateDistanceToBall(pPlayer), 1.0, 9.0, FALSE);
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Teammate-Has-Ball  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if a teammate is in front of me in x direction (used for offside prevention)
-----------------------------------------------------------------------------*/
float GetTeammateInFront()			
{
    static float val = 0.0;
    static float fTeammateX;
    Unum pPlayer = Mem->FurthestForwardTeammate();

    if (pPlayer != Unum_Unknown)
 {
    	fTeammateX = Mem->TeammateX(pPlayer) - Mem->MyX();
	val = GetLinearFuzzyValue(fTeammateX, 2.0, 5.0);
    }
    else
	val = 0.0;

    Mem->LogAction3(200,"Precp Teammate-InFront  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
prevents the goalie from running to the ball if a teammate is between him and the ball.
The goalie has to run to the ball even if he is not the closest teammate to the ball.
This is the case, if an opponent followed by a teammate is running to the goal
-----------------------------------------------------------------------------*/
float GetTeammateIsBetweenBall()		
{
    static float val = 0.0;
    static float fNearer, deltaDist;
    
    fNearer = GetTeammateIsNearerBall();
    
    if (fNearer > 0.5)
 {
    	Unum	ownPlayer = Mem->ClosestTeammateToBall();
	Unum	opponent  = Mem->ClosestOpponentToBall();
	Vector  ownGoal   = Mem->MarkerPosition( Mem->RM_My_Goal );

	if ((ownPlayer == Unum_Unknown) || (opponent == Unum_Unknown))
 val =  fNearer;
	else {
	    deltaDist = Mem->OpponentDistanceTo(opponent,ownGoal) - 
		    Mem->TeammateDistanceTo(ownPlayer,ownGoal);

	    val = GetLinearFuzzyValue(deltaDist, 1.0, 3.0) * fNearer;
	}
    }
    else
	val = fNearer;

    Mem->LogAction3(200,"Precp Teammate-Is-Between-Ball  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
true if one of the teammates is unmarked
-----------------------------------------------------------------------------*/
float GetTeammateIsFree()
{
    static float val = 0.0;


    if ((Mem->FreestTeammate = CalculateFreestTeammate()) == Unum_Unknown)
 
    	val = 0.0;
    else
	val = Mem->PassDegree; 

    Mem->LogAction3(200,"Precp Teammate-Is-Free  \t\t%1.2f", val);
    return val;

}

/*-----------------------------------------------------------------------------
true if a teammate is closer to the ball than this player
-----------------------------------------------------------------------------*/
float GetTeammateIsNearerBall()				
{
    static float val = 0.0;

    Unum	ownPlayer = Mem->ClosestTeammateToBall();
    float	deltaDist;
    Vector	ballInFuture;

    if((Mem->CP_goalie && Mem->PlayMode == PM_Play_On) ||
	ownPlayer == Unum_Unknown || ownPlayer == 1 || ownPlayer == Mem->MyNumber)
	val = 0.0;
    else if(Mem->AmIMid() &&
              Mem->TeammateWing(ownPlayer) == Mem->MyWing() && Mem->IsTeammateDef(ownPlayer))
	    val = 0.0; /* covering defence line */
    else if(Mem->PlayerWithBall() == ownPlayer)
	    val = 1.0;
    else {
/*
	Unum player = Mem->BallPossessor();
	if(player == Mem->MyNumber || player <= 0)
	    val = 0;
	else
	    val = GetHaveSeenBall();
*/
	ballInFuture = Mem->BallPredictedPosition( 7 );
    deltaDist = Mem->DistanceTo(ballInFuture) - 
	        Mem->TeammateDistanceTo(ownPlayer, ballInFuture);

    val = GetLinearFuzzyValue(deltaDist, 0.0, 1.0) * GetHaveSeenBall();
    }
    
    Mem->LogAction3(200,"Precp Teammate-Is-Nearer-Ball  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
Shows when goalie start
s
-----------------------------------------------------------------------------*/
float GetTeammatesInPlaces()
{
    static float val = 0.0;

    if(Mem->PlayMode != PM_My_Goalie_Free_Kick)	val = 0.0;
    else if(Mem->CurrentTime.t - Mem->LastCatchTime.t > 50) val = 1.0;
    else val = 0.0;

    Mem->LogAction3(200,"Precp Teammates-In-Places  \t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
Shows when to get an offside trap.
-----------------------------------------------------------------------------*/

float GetOffsidePosition()			
{
    static float val = 0.0;

    val =  Max(GetBallAhead(), (1.0 - GetLowStamina()));

    Mem->LogAction3(200,"Precp Offside-Position  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
True if it's my goal-kick now
-----------------------------------------------------------------------------*/
float GetMyGoalKick()
{
    static float val = 0.0;

    if(Mem->PlayMode == PM_My_Goalie_Free_Kick) val = 1.0;
    else val = 0.0;

    Mem->LogAction3(200,"Precp My-Goal-Kick  \t\t%1.2f", val);
    return val;
}

/*-----------------------------------------------------------------------------
True if it's there goal-kick now
-----------------------------------------------------------------------------*/
float GetTheirGoalKick()
{
    static float val = 0.0;

    if(Mem->PlayMode == PM_Their_Goalie_Free_Kick) val = 1.0;
    else val = 0.0;

    Mem->LogAction3(200,"Precp Their-Goal-Kick  \t%1.2f", val);
    return val;
}

	
	